import _sortBy from "lodash/sortBy";
import _each from "lodash/each";
import _get from "lodash/get";
import _transform from "lodash/transform";
import _trim from "lodash/trim";
import _every from "lodash/every";
import _map from "lodash/map";

function _typeof(obj) { if (typeof Symbol === "function" && typeof Symbol.iterator === "symbol") { _typeof = function _typeof(obj) { return typeof obj; }; } else { _typeof = function _typeof(obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; }; } return _typeof(obj); }

import searchTextTokenizer from 'search-text-tokenizer';

function indexOfTerm(term, value, valueStr) {
  if (!value) return -1;

  var type = _typeof(value);

  if (type === 'string') {
    return value.toLowerCase().indexOf(term);
  } else if (type === 'number' && !isNaN(term)) {
    var termWithDot = term.indexOf('.');

    if (termWithDot < 0) {
      // match as integer
      if (Math.trunc(value) === parseInt(term, 10)) {
        return 0;
      }
    } else if (valueStr === term) {
      // float, full match
      return 0;
    }
  }

  return -1;
}

function matchTerm(item, term) {
  // eslint-disable-next-line
  for (var field in item) {
    var value = item[field];
    var valueStr = "".concat(value);
    var index = indexOfTerm(term, value, valueStr); // console.log('>> matchTerm >>', index, value, term)

    if (index >= 0) {
      return {
        term: term,
        field: field,
        index: index,
        value: value,
        percentage: term.length / valueStr.length
      };
    }
  }

  return null;
}

function matchKeywords(item, keywords) {
  var matches = _map(keywords, function (keyword) {
    var term = keyword.term,
        exclude = keyword.exclude;
    var fieldIndex = matchTerm(item, term);

    if (exclude) {
      if (fieldIndex) return null;
      return {
        term: term,
        exclude: true,
        field: 'exclude',
        index: 0,
        percentage: 0.5
      };
    }

    return fieldIndex;
  });

  return _every(matches) ? matches : null;
}

function calcRelevance(matches) {
  var sort = 1000;

  for (var i = 0, ii = matches.length; i < ii; i++) {
    var match = matches[i];

    if (i > 0) {
      var lastMatch = matches[i - 1]; // same field and after last index

      if (match.field === lastMatch.field) {
        // same field, follow sequence of keywords
        if (match.index >= lastMatch.index + lastMatch.term.length) {
          sort -= 5 * match.percentage;
          break;
        } // same field, before last index


        if (match.index !== lastMatch.index) {
          sort -= 3 * match.percentage;
          break;
        }
      }
    } // diff field or first field


    sort -= match.percentage;
  }

  return sort;
}

export function tokenizeKeywords(searchStr) {
  if (!searchStr) return null; // can use tag which extracted from searchTextTokenizer in future
  // '-car' => [ { term: 'car', exclude: true } ]

  var keywords = searchTextTokenizer(searchStr);
  if (keywords.length === 0) return null;
  return _map(keywords, function (keyword) {
    if (keyword.tag) {
      // kind of remove tag feature
      keyword.term = "".concat(keyword.tag, ":").concat(keyword.term);
    }

    return keyword;
  });
}

var defaultPickFields = function defaultPickFields(doc) {
  return doc;
};

var defaultTuneOrder = function defaultTuneOrder(_order) {
  return _order;
};

export default function searchObjects(docs, search, conf) {
  var keywords = Array.isArray(search) ? search : tokenizeKeywords(_trim(search).toLowerCase());
  if (!keywords || keywords.length === 0) return docs;
  var hasExclude = false;

  var terms = _map(keywords, function (keyword) {
    if (keyword.exclude) hasExclude = true;
    return keyword.term;
  });

  var noExcludeSearchStr = hasExclude ? null : terms.join(' '); // normalize options

  var optType = _typeof(conf);

  var pick;
  var tuneOrder;

  if (optType === 'function' || Array.isArray(conf)) {
    pick = conf;
  } else if (optType === 'object') {
    pick = conf.pick;
    tuneOrder = conf.tuneOrder;
  }

  if (!pick) {
    pick = defaultPickFields;
  } else if (Array.isArray(pick)) {
    var fields = pick;

    pick = function pick(doc) {
      return _transform(fields, function (ret, field) {
        ret[field] = _get(doc, field);
      }, {});
    };
  }

  if (!tuneOrder) tuneOrder = defaultTuneOrder; // filter

  var items = [];

  _each(docs, function (doc) {
    var item = _typeof(doc) === 'object' ? pick(doc) : {
      value: doc
    };

    if (noExcludeSearchStr) {
      var wholeMatch = matchTerm(item, noExcludeSearchStr);

      if (wholeMatch) {
        item._order = tuneOrder(1 - wholeMatch.percentage, doc);
        item._doc = doc;
        items.push(item);
        return;
      }
    }

    var matches = matchKeywords(item, keywords);

    if (matches) {
      item._order = tuneOrder(calcRelevance(matches), doc);
      item._doc = doc;
      items.push(item);
    }
  }); // _order should be float between 0 to 1 and sort in asc
  // sort


  return _map(_sortBy(items, '_order'), '_doc');
}